// defines
#define G_SRATE 44100
#define FRAMESIZE 512
#define NCHANNELS  2
#define NNOTES 1
#define NPASTFREQS 5

#define FORMAT RTAUDIO_FLOAT64

#define MODE_NORMAL 0
#define MODE_MOVECIRCLE 1
#define MODE_ENTERWORD 2
#define MODE_ENTERWORDS 3

#define VIEW_SLEWRATE 5
#define VIEW_SLEWRATE_SLOW 1

#define SLEWRATE_WORDS 1

#define G_PI 3.14159

#ifndef __W_GLOBALS_H__
#define __W_GLOBALS_H__

#include "x-def.h"
#include "x-audio.h"
#include "x-gfx.h"
#include "x-vector3d.h"
#include "y-waveform.h"

#include "words.h"
#include "RtMidi.h"
#include "RtAudio.h"

// c++
#include <string>
#include <map>
#include <vector>
#include <utility>

#include <iostream>
#include <fstream>

using namespace std;

// forward reference
class Sim;
class wsWord;
class Words;
class Pane;

class wsSynth;

//-----------------------------------------------------------------------------
// name: class Globals
// desc: the global class
//-----------------------------------------------------------------------------
class Globals
{
public:
    // top level root simulation
    static Sim * sim;

    // version
    static std::string version;
    
    // width and height of the window
    static GLsizei windowWidth;
    static GLsizei windowHeight;
    static GLsizei lastWindowWidth;
    static GLsizei lastWindowHeight;
    
    // graphics fullscreen
    static GLboolean fullscreen;
    // blend pane instead of clearing screen
    static GLboolean blendScreen;
    // blend screen parameters
    static Vector3D blendAlpha;
    static GLfloat blendRed;
    // fill mode
    static GLenum fillmode;
    // background color
    static iSlew3D bgColor;
    // view stuff
    static Vector3D viewRadius;
    static Vector3D viewEyeY;
    static Vector3D viewEyeX;
    static Vector3D fov;
    static Vector3D viewX;
    static Vector3D viewY;
    static Vector3D viewZ;
    
    // light 0 position
    static GLfloat light0_pos[4];
    // light 1 parameters
    static GLfloat light1_ambient[4];
    static GLfloat light1_diffuse[4];
    static GLfloat light1_specular[4];
    static GLfloat light1_pos[4];
    
    // words
    static Words * wordList;
    static string wordFile, tempFile, pastFreqsFile;
    
    // PCA
    static string pcaFile;
    
    // pane
    static Pane * controlPanel;
    static Vector3D txtEnterWordPos, txtRatePos, txtSimilarWordsPos, txtRatingPos;
    static Vector3D txtEnterWordCol, txtRateCol, txtSimilarWordsCol, txtRatingCol;
    static Vector3D raterLineStart, raterLineEnd;
    static Vector3D raterCircleColor, raterLineColor;
    static float raterCircleRadius;
    static Vector3D wordBoxCorner1, wordBoxCorner2;
    static Vector3D wordsBoxCorner1, wordsBoxCorner2;
    
    // interaction mode
    static int interactionMode;
    
    // focal words
    static wsWord * focalWord;
    static wsWord * ringingWord;
    static Vector3D * focalWordTrans;
    
    // midi
    static RtMidiIn * midiIn;
    
    // audio
    static RtAudio * audioIO;
    static RtAudio::StreamOptions * ioStreamOptions;
    static RtAudio::StreamParameters * iStreamParams;
    static RtAudio::StreamParameters * oStreamParams;
    
    static unsigned int bufferBytes;
    static unsigned int bufferFrames;
    
    // synthesizers
    static vector<wsSynth *> synth;
    static int * synthNote;
    static int synthInd;
    
    static int inPlay;
    
    // randomness
    static GLfloat * pastFreqs;
    static int pastFreqsInd;
    static int isRinging;
    static GLfloat timeSinceLastRing;
    
    // time
    static GLfloat timeSinceLastR;
    static int hasExploded;
    static GLfloat timeSinceLastInteraction;
    
    static GLfloat eighthNoteTempo;
    static GLfloat noteDuration;
    static GLfloat noteDurationHold;
    static GLfloat noteDurationHoldTypical;
    
};



#endif

