#include "pane.h"
#include "globals.h"
#include <iostream>

using namespace std;

void Pane::initialize(){
    
    // y Translation
    yTrans.value = 0;
    yTrans.update(yTransFinal, yTransSlewRate);
    
    // Text
    txtEnterWord.setString("Enter Word:");
    txtRate.setString("Rate Correspondence:");
    txtSimilarWords.setString("What word(s) does this sound like?");
    txtRating.setString(to_string(rater.getCirclePos()));
    
    txtEnterWord.setPos(Globals::txtEnterWordPos);
    txtRate.setPos(Globals::txtRatePos);
    txtSimilarWords.setPos(Globals::txtSimilarWordsPos);
    txtRating.setPos(Globals::txtRatingPos);
    
    txtEnterWord.setCol(Globals::txtEnterWordCol);
    txtRate.setCol(Globals::txtRateCol);
    txtSimilarWords.setCol(Globals::txtSimilarWordsCol);
    txtRating.setCol(Globals::txtRatingCol);
    
    // rating Line
    rater.setCircleColor(Globals::raterCircleColor);
    rater.setLineColor(Globals::raterLineColor);
    rater.setLinePos(Globals::raterLineStart, Globals::raterLineEnd);
    rater.setCircleRadius(Globals::raterCircleRadius);
    rater.setCirclePos(0);
    
    // text boxes
    word.setBoxCorners(Globals::wordBoxCorner1, Globals::wordBoxCorner2);
    words.setBoxCorners(Globals::wordsBoxCorner1, Globals::wordsBoxCorner2);
    word.addLetter("");
    words.addLetter("");
    
};

// render control pane
void Pane::render(){
    
    // enable state
    glEnableClientState( GL_VERTEX_ARRAY );
    glEnableClientState( GL_NORMAL_ARRAY );
    
    // push
    glPushMatrix();
    
    // 2D View
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_CULL_FACE);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(100, 100, 0, 100);
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glColor3f(1, 1, 1);
    
    // Main Console
    glEnable( GL_BLEND );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

    
    glBegin(GL_QUADS);
    glColor4d(1,1,1,0);
    glVertex2f(panePos1[0],panePos1[1] + yTrans.value);
    glColor4d(1,1,1,.3);
    glVertex2f(panePos1[0],panePos2[1] + yTrans.value);
    glColor4d(1,1,1,.5);
    glVertex2f(panePos2[0],panePos2[1] + yTrans.value);
    glColor4d(1,1,1,0);
    glVertex2f(panePos2[0],panePos1[1] + yTrans.value);
    glEnd();
    
    // Text
    txtEnterWord.render(0, yTrans.value);
    txtRate.render(0, yTrans.value);
    txtSimilarWords.render(0, yTrans.value);
    txtRating.setString(to_string(rater.getCirclePos()));
    txtRating.render(0, yTrans.value);
    
    // Line
    rater.render(0, yTrans.value);
    
    // Text boxes
    word.render(0, yTrans.value);
    words.render(0, yTrans.value);
    
    // pop
    glPopMatrix();
    
    // disable
    glDisableClientState( GL_VERTEX_ARRAY );
    glDisableClientState( GL_NORMAL_ARRAY );

}

void Pane::slideUp(){
    yTrans.update(0,yTransSlewRate);
    
}
void Pane::slideDown(){
    yTrans.update(yTransFinal,yTransSlewRate);
};

int Pane::getYTrans(){
    return yTrans.value;
};

void Pane::updateAll(YTimeInterval dt){
    yTrans.interp(dt);
};

void RatingLine::render(float x, float y){
    Vector3D circleCenter = Vector3D(posStart[0]+circlePos*(posEnd[0]-posStart[0])+x,posStart[1]+circlePos*(posEnd[1]-posStart[1])+y,posStart[2]+circlePos*(posEnd[2]-posStart[2]));
    
    glColor4d(colorLine[0],colorLine[1],colorLine[2],1);
    glBegin(GL_LINES);
    glVertex2f(posStart[0],posStart[1]+y);
    glVertex2f(posEnd[0],posEnd[1]+y);
    glEnd();
    
    glColor4d(colorLine[0],colorLine[1],colorLine[2],1);
    glBegin(GL_LINES);
    glVertex2f(posStart[0],posStart[1]+y);
    glVertex2f(posEnd[0],posEnd[1]+y);
    glEnd();
    
    GLUquadricObj *disk;
    disk = gluNewQuadric();
    
    glPushMatrix();
    glTranslatef(circleCenter[0],circleCenter[1],0);
    glColor4d(colorCircle[0],colorCircle[1],colorCircle[2],1);
    gluDisk(disk, 0, rCircle, 10,20);
    glPopMatrix();
    
}

void TextBox::render(float x, float y){
    
    glColor4d(1,1,1,.75);
    glBegin(GL_QUADS);
    glVertex2f(boxCorner1[0]+x,boxCorner1[1]+y);
    glVertex2f(boxCorner1[0]+x,boxCorner2[1]+y);
    glVertex2f(boxCorner2[0]+x,boxCorner2[1]+y);
    glVertex2f(boxCorner2[0]+x,boxCorner1[1]+y);
    glEnd();
    
    glColor4d(0,0,0,1);
    glRasterPos2f(boxCorner1[0] + x + .01,boxCorner1[1] + y - .03);
    for (float i = 0; i < text.length(); i++){
        glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18,text[i]);
    }
    
}