#ifndef __W_PANE_H__
#define __W_PANE_H__

#include "y-entity.h"
#include <vector>
#include <iostream>
#include <algorithm>

#include "globals.h"

#define yTransFinal -.3
#define yTransSlewRate 5

using namespace std;

// forward reference
class Globals;

// 2D text
class Text2D {
    
private:
    
    string text;
    Vector3D position;
    Vector3D color;
    
public:
    
    Text2D(){};
    
    ~Text2D(){};
    
    void render(GLfloat x, GLfloat y){
        glColor4d(color[0],color[1],color[2],1);
        glRasterPos2f(position[0] + x,position[1] + y);
        for (float i = 0; i < text.length(); i++){
            glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_24,text[i]);
        }
    };
    
    void setString(string txt){ text = txt.c_str(); };
    void setPos(Vector3D pos){ position = Vector3D(pos[0],pos[1],pos[2]); };
    void setCol(Vector3D col){ color = Vector3D(col[0],col[1],col[2]); };
    
    int getLen(){return (int) text.length();};
    string getString(){return text.c_str(); };
    
};

// Rating Line
class RatingLine {
    
private:
    
    Vector3D posStart, posEnd, colorLine, colorCircle;
    float rCircle, circlePos;
    
public:
    
    RatingLine(){};
    ~RatingLine(){};
    
    void setLinePos(Vector3D pStart, Vector3D pEnd){ posStart = Vector3D(pStart[0], pStart[1], pStart[2]); posEnd = Vector3D(pEnd[0], pEnd[1], pEnd[2]);};
    void setCirclePos(float x){circlePos = x;};
    void setCirclePos(float x, float y){
        
        // hack (we assume a horizontal line!)
        float x_ = (x-posStart[0])/(posEnd[0]-posStart[0]);
        if (x_ > 1) setCirclePos(1);
        else if (x_ < 0) setCirclePos(0);
        else setCirclePos(x_);
        
    }
    void setLineColor(Vector3D col){colorLine = Vector3D(col[0],col[1],col[2]);};
    void setCircleColor(Vector3D col){colorCircle = Vector3D(col[0],col[1],col[2]);};
    void setCircleRadius(float x){rCircle = x;};
    
    float getCirclePos(){return circlePos;};
    
    int isInArea(float mouseX, float mouseY, float transX, float transY){
        if ( (mouseX <= posEnd[0]+transX) && (mouseX >= posStart[0]+transX) && (mouseY <= posEnd[1]+rCircle+transY) && (mouseY > posEnd[1]-rCircle+transY) ) return 1;
        else return 0;
    }
    
    void render(float x, float y);
    
};

// Text Box
class TextBox {
    
private:
    string text;
    Vector3D boxCorner1, boxCorner2;
    
public:
    
    TextBox(){};
    ~TextBox(){};
    
    void initialize(){};
    
    void addLetter(string letter){ text.append(letter);  };
    void removeLetter(){if (text.length()> 0) text = text.substr(0,text.length()-1);};
    void setBoxCorners(Vector3D c1, Vector3D c2){ boxCorner1 = Vector3D(c1[0],c1[1],c1[2]); boxCorner2 = Vector3D(c2[0],c2[1],c2[2]);};
    
    void render(float x, float y);
    
    int isInArea(float mouseX, float mouseY, float transX, float transY){
        if ( (mouseX < boxCorner2[0]+transX) && (mouseX > boxCorner1[0]+transX) && (mouseY < boxCorner1[1]+transY) && (mouseY > boxCorner2[1]+transY) ) return 1;
        else return 0;
    }
    
    string getString(){return text.c_str();};
    
};

// Pane
class Pane : public YEntity{

private:
    
    Vector3D panePos1 = Vector3D(-1, -.7, 0);
    Vector3D panePos2 = Vector3D(1, -1, 0);
    Vector3D yTrans = Vector3D(0,yTransFinal,0);
    
    Text2D txtEnterWord, txtRate, txtSimilarWords, txtRating;
    
    RatingLine rater;
    
    TextBox word, words;
    
public:
    
    Pane(){}
    ~Pane(){}
    
    void initialize();
    
    void render();
    
    void slideUp();
    void slideDown();
    
    int getYTrans();
    string getWord(){return word.getString();};
    string getWords(){return words.getString();};
    int getWordsLength(){return words.getString().length();};
    
    int isInCircle(float x, float y){return rater.isInArea(x,y,0,yTrans.value);};
    int isInWord(float x, float y){return word.isInArea(x,y,0,yTrans.value);};
    int isInWords(float x, float y){return words.isInArea(x,y,0,yTrans.value);};
    
    int isInObject(float x, float y){
        if (isInCircle(x,y)==1) {
            return MODE_MOVECIRCLE;
        }
        else if (isInWord(x,y)==1){
            return MODE_ENTERWORD;
        }
        else if (isInWords(x,y)==1){
            return MODE_ENTERWORDS;
        }
        else return MODE_NORMAL;
    }
    
    void write(int mode, unsigned char letter){
        
        // if it's a leter, make it lowercase
        if (isalpha(letter)){ if( isupper(letter) ) letter +=32;}
        // otherwise, make it a comma
        else{ letter = ',';}
        
        // convert to string
        string str(1, letter);
        
        switch (mode){
            case (MODE_ENTERWORD): if (isalpha(letter)) word.addLetter(str); break;
            case (MODE_ENTERWORDS): words.addLetter(str); break;
        }
        
        
    }
    
    void erase(int mode){
        switch (mode){
            case (MODE_ENTERWORD): word.removeLetter(); break;
            case (MODE_ENTERWORDS): words.removeLetter(); break;
        }
    }
    
    void updateAll(YTimeInterval dt);
    
    void setCirclePos(float x, float y){ rater.setCirclePos(x, y);};
    void setCirclePos(float val){ rater.setCirclePos(val); };
    float getCirclePos(){ return rater.getCirclePos(); };
    
};

#endif