#ifndef __W_SYNTH_H__
#define __W_SYNTH_H__

#include <vector>
#include <ioStream>

#include "Stk.h"
#include "BandedWG.h"
#include "SineWave.h"
#include "BlitSquare.h"
#include "ADSR.h"
#include "FormSwep.h"
#include "FMVoices.h"
#include "BiQuad.h"
#include "Echo.h"

#include "globals.h"

using namespace std;
using namespace stk;

// forward reference
class Globals;
class wsSynthSettings;

class wsSynth{
    
public:
    
    // sources
    SineWave sw;
    BlitSquare sq;
    FMVoices fm;
    
    // effects
    ADSR adsrGain, adsrFreq, adsrRad;
    FormSwep formSwep;
    BiQuad filter;
    Echo echo;
    
    wsSynth(){};
    ~wsSynth(){};
    
    void initialize(){
        
        // gain
        adsrGain.setSampleRate(G_SRATE);
        adsrGain.setAttackTarget(1);
        adsrGain.setSustainLevel(.3);
        adsrGain.setAttackTime(.2);
        adsrGain.setDecayTime(.2);
        adsrGain.setReleaseTime(1);

        // frequency sweep (adsr)
        adsrFreq.setSampleRate(G_SRATE);
        adsrFreq.setAttackTarget(800.2);
        adsrFreq.setSustainLevel(21.3);
        adsrFreq.setAttackTime(.2);
        adsrFreq.setDecayTime(.2);
        adsrGain.setReleaseTime(1);
        
        // frequency radius sweep (adsr)
        adsrRad.setSampleRate(G_SRATE);
        adsrRad.setAttackTarget(.9);
        adsrRad.setSustainLevel(.3);
        adsrRad.setAttackTime(.2);
        adsrRad.setDecayTime(.2);
        adsrRad.setReleaseTime(1);
        
        // square, sine waves
        sq.setSampleRate(G_SRATE);
        sw.setSampleRate(G_SRATE);
        
        // LFO
        formSwep.setSampleRate(G_SRATE);
        formSwep.setStates(100,.5,1);
        formSwep.setTargets(1000,.5,1);
        formSwep.setSweepRate(.2);
        
        // fm synthesizer
        fm.setSampleRate(G_SRATE);
        fm.setModulationDepth(.5);
        fm.setGain(1,0);
        fm.setGain(2,0);
        fm.setRatio(3,.5);
        fm.setGain(3,.02);
        
        // biquad
        filter.setResonance(100, .9);
        
        // delay
        echo.setMaximumDelay((unsigned long) SRATE*echoMax);
        echo.setDelay((unsigned long) SRATE/2);
        echo.setEffectMix(.2);
        
    };
    
};

#endif