#include "words.h"

#include "y-entity.h"

#include <fstream>
#include <iomanip>
#include <iostream>
#include <string>

#define STRING 0
#define POS_X 1
#define POS_Y 2
#define POS_Z 3
#define COL_X 4
#define COL_Y 5
#define COL_Z 6
#define FMRATIO_0 7
#define FMRATIO_1 8
#define FMRATIO_2 9
#define FMRATIO_3 10
#define FMGAIN_0 11
#define FMGAIN_1 12
#define FMGAIN_2 13
#define FMGAIN_3 14
#define ADSRTIME_0 15
#define ADSRTIME_1 16
#define ADSRTIME_2 17
#define ASGAIN_0 18
#define ASGAIN_1 19
#define ASFREQ_0 20
#define ASFREQ_1 21
#define ASRAD_0 22
#define ASRAD_1 23
#define ECHO 24

#define NOTE_0 25
#define NOTE_1 26
#define NOTE_2 27
#define NOTE_3 28
#define NOTE_4 29
#define TEMPO 30

#define CONF 31
#define NVOTES 32

#define PARAMS_SCALE_ASRAD_1 10
#define PARAMS_SCALE_FMGAIN_3 10
#define PARAMS_SCALE_FMRATIO_3 10

#define PCASCALE_0 150
#define PCASCALE_1 150
#define PCASCALE_2 20

using namespace std;

// -----------------------
// initialize class: Words
// -----------------------
void Words::initialize(string fileName){
 
    int iChar, iParam;      // indices
    string myLine, myWord;  // strings
    char myChar;            // character (for comparison purposes)
    
    Vector3D pos, col;      // future word positions
    
    ifstream myFile;        // file token
    myFile.open(fileName);  // open file
    
    cerr << "reading file ..." << endl;

    while(!myFile.eof())
    {
        wsWord * newWord = new wsWord();    // wsWord class [contains string, character, etc.]
        float conf = 0;
        
        getline(myFile, myLine);    // get next line
        
        iParam = 0;             // reset parameter counter
        iChar = 0;              // reset character counter
        pos = Vector3D(0,0,0);  // reset position
        col = Vector3D(1,1,1);  // reset color
        myWord.clear();
        
        if (myLine.size() > 0){
            while( (iChar < myLine.size()) && (iParam < NPARAMS) )
            {
            
                myChar = myLine.at(iChar);
                if (myChar == ',' || myChar == '\n')
                {
                    switch (iParam){
                        case STRING:
                            newWord->setString(myWord, iChar);
                            break;
                        case POS_X:
                            pos.x = atof(myWord.c_str());
                            break;
                        case POS_Y:
                            pos.y = atof(myWord.c_str());
                            break;
                        case POS_Z:
                            pos.z = atof(myWord.c_str());
                            break;
                        case COL_X:
                            col.x = atof(myWord.c_str());
                            break;
                        case COL_Y:
                            col.y = atof(myWord.c_str());
                            break;
                        case COL_Z:
                            col.z = atof(myWord.c_str());
                            break;
                        case FMRATIO_0:
                            newWord->params[FMRATIO_0-7] = (double) atof(myWord.c_str());
                            newWord->synthSettings->setFMRatios(0, (StkFloat) atof(myWord.c_str()));
                            break;
                        case FMRATIO_1:
                            newWord->params[FMRATIO_1-7] = (double) atof(myWord.c_str());
                            newWord->synthSettings->setFMRatios(1, (StkFloat) atof(myWord.c_str()));
                            break;
                        case FMRATIO_2:
                            newWord->params[FMRATIO_2-7] = (double) atof(myWord.c_str());
                            newWord->synthSettings->setFMRatios(2, (StkFloat) atof(myWord.c_str()));
                            break;
                        case FMRATIO_3:
                            newWord->params[FMRATIO_3-7] = (double) PARAMS_SCALE_FMRATIO_3*atof(myWord.c_str());
                            newWord->synthSettings->setFMRatios(3, (StkFloat) atof(myWord.c_str()));
                            break;
                        case FMGAIN_0:
                            newWord->params[FMGAIN_0-7] = (double) atof(myWord.c_str());
                            newWord->synthSettings->setFMGains(0, (StkFloat) atof(myWord.c_str()));
                            break;
                        case FMGAIN_1:
                            newWord->params[FMGAIN_1-7] = (double) atof(myWord.c_str());
                            newWord->synthSettings->setFMGains(1, (StkFloat) atof(myWord.c_str()));
                            break;
                        case FMGAIN_2:
                            newWord->params[FMGAIN_2-7] = (double) atof(myWord.c_str());
                            newWord->synthSettings->setFMGains(2, (StkFloat) atof(myWord.c_str()));
                            break;
                        case FMGAIN_3:
                            newWord->params[FMGAIN_3-7] = (double) PARAMS_SCALE_FMRATIO_3*atof(myWord.c_str());
                            newWord->synthSettings->setFMGains(3, (StkFloat) atof(myWord.c_str()));
                            break;
                        case ADSRTIME_0:
                            newWord->params[ADSRTIME_0-7] = (double) atof(myWord.c_str());
                            newWord->synthSettings->setADSRTimes(0,(StkFloat) atof(myWord.c_str()));
                            break;
                        case ADSRTIME_1:
                            newWord->params[ADSRTIME_1-7] = (double) atof(myWord.c_str());
                            newWord->synthSettings->setADSRTimes(1,(StkFloat) atof(myWord.c_str()));
                            break;
                        case ADSRTIME_2:
                            newWord->params[ADSRTIME_2-7] = (double) atof(myWord.c_str());
                            newWord->synthSettings->setADSRTimes(2,(StkFloat) atof(myWord.c_str()));
                            break;
                        case ASGAIN_0:
                            newWord->params[ASGAIN_0-7] = (double) atof(myWord.c_str());
                            newWord->synthSettings->setASGain(0, (StkFloat) atof(myWord.c_str()));
                            break;
                        case ASGAIN_1:
                            newWord->params[ASGAIN_1-7] = (double) atof(myWord.c_str());
                            newWord->synthSettings->setASGain(1, (StkFloat) atof(myWord.c_str()));
                            break;
                        case ASFREQ_0:
                            newWord->params[ASFREQ_0-7] = (double) atof(myWord.c_str());
                            newWord->synthSettings->setASFreq(0, (StkFloat) atof(myWord.c_str()));
                            break;
                        case ASFREQ_1:
                            newWord->params[ASFREQ_1-7] = (double) atof(myWord.c_str());
                            newWord->synthSettings->setASFreq(1, (StkFloat) atof(myWord.c_str()));
                            break;
                        case ASRAD_0:
                            newWord->params[ASRAD_0-7] = (double) atof(myWord.c_str());
                            newWord->synthSettings->setASRad(0, (StkFloat) atof(myWord.c_str()));
                            break;
                        case ASRAD_1:
                            newWord->params[ASRAD_1-7] = (double) PARAMS_SCALE_ASRAD_1*atof(myWord.c_str());
                            newWord->synthSettings->setASRad(1, (StkFloat) atof(myWord.c_str()));
                            break;
                        case ECHO:
                            newWord->params[ECHO-7] = (double) atof(myWord.c_str());
                            newWord->synthSettings->setEcho((unsigned long) atoi(myWord.c_str()));
                            break;
                        case CONF:
                            conf = atof(myWord.c_str());
                            break;
                        case NVOTES:
                            newWord->resetConfidence(conf, atoi(myWord.c_str()));
                            break;
                        case NOTE_0:
                            newWord->setFreqs(0, atof(myWord.c_str()));
                            break;
                        case NOTE_1:
                            newWord->setFreqs(1, atof(myWord.c_str()));
                            break;
                        case NOTE_2:
                            newWord->setFreqs(2, atof(myWord.c_str()));
                            break;
                        case NOTE_3:
                            newWord->setFreqs(3, atof(myWord.c_str()));
                            break;
                        case NOTE_4:
                            newWord->setFreqs(4, atof(myWord.c_str()));
                            break;
                        case TEMPO:
                            newWord->setTempo(atof(myWord.c_str()));
                            break;
                    }
                
                    myWord.clear();
                    iParam++;
                }
                else myWord.append(myLine.substr(iChar,1));
            
                iChar++;
            }
            
            if(iParam==NPARAMS-1) newWord->resetConfidence(conf, atoi(myWord.c_str()));

            newWord->setPosFinal(pos);
            newWord->setPosGoal();
            newWord->updatePos();
            newWord->setColor (col);
            
            // hash it!
            existingWordsMap.insert( {{newWord->getString(), newWord}} );
            
            nWords = nWords + 1;
            
        }
        
    }
    
}

void Words::addList(string str, wsWord * baseWord){
    
    string word = "";
    wsWord * theWord;
    Vector3D * vec = new Vector3D();
    int isListed;
    
    if (str.size() > 0){

        str = str.append(",");
        
        for (int iChar = 0; iChar < str.size(); iChar++){
            if (str[iChar] != ',') word = word.append(str.substr(iChar, 1));
            else{
                
                isListed = Globals::wordList->isListed(word);
                
                
                switch (isListed){
                    case 0:
                    {
                        // create a new word
                        theWord = new wsWord();
                        theWord->setString(word, (int) word.size());
                        
                        // set its position
                        Globals::focalWord->getPosFinal(vec);
                        theWord->setPosFinal(*vec);
                        theWord->setPosGoal();
                        theWord->yText->setCenterLocation(*vec);
                        
                        // set its color
                        Globals::focalWord->getCol(vec);
                        theWord->setColor(*vec);
                        theWord->yText->setColor(Vector3D((float)rand()/RAND_MAX,(float)rand()/RAND_MAX,(float)rand()/RAND_MAX));
                        
                        // copy over synthesizer properties
                        theWord->setSynth(Globals::focalWord->synthSettings);
                        
                        // update parameters
                        theWord->setParams(Globals::focalWord->params);
                        
                        // set confidence level to .2 given 1 vote
                        theWord->resetConfidence(.6, 1);
                        
                        // add new word to simulation
                        Globals::wordList->newWordsMap.insert( {{theWord->getString(), theWord}} );
                        Globals::sim->root().addChild(theWord->yText);
                        
                        nWords = nWords + 1;
                        
                        break;
                    }
                    case (1):
                    {
                        // get word of interest
                        theWord = Globals::wordList->existingWordsMap.find(word)->second;
                        
                        // move from existingWordsMap to editedWordsMap
                        Globals::wordList->existingWordsMap.erase(word);
                        Globals::wordList->editedWordsMap.insert({{word, theWord}});
                        
                        // modify confidence & synthesizer properties
                        theWord->updateConfidence(Globals::focalWord->synthSettings, .6);
                        
                        break;
                        
                    }
                    case (2):
                    {
                        // get word of interest
                        theWord = Globals::wordList->newWordsMap.find(word)->second;
                        
                        // modify confidence & synthesizer properties
                        theWord->updateConfidence(Globals::focalWord->synthSettings, .6);
                        
                        break;
                    }
                    case(3):
                    {
                        // change focal point of word list
                        Globals::focalWord = Globals::wordList->editedWordsMap.find(word)->second;
                        // modify confidence & synthesizer properties
                        theWord->updateConfidence(Globals::focalWord->synthSettings, .6);
                        break;
                    }
                        
                }
                
                word.clear();
                
            }
        }
    
    }
    
    
}

// ----------
// save words
// ----------
void Words::save(string fileName){
    
    ofstream myOut;
    myOut.open(Globals::tempFile);
    string myLine, myWord;
    
    // save all existing words
    for (auto& a: existingWordsMap) saveLine(&myOut, a.second);
    
    // save all edited words
    for (auto& a: editedWordsMap) saveLine(&myOut, a.second);
    
    // save all the new words
    for (auto& a: newWordsMap) saveLine(&myOut, a.second);
    
    myOut.close();
    
    // move temporary file to actual file
    rename( Globals::tempFile.c_str(), Globals::wordFile.c_str() );
    
};

// -------------------------
// write line to a save file
// -------------------------
void Words::saveLine(ofstream * myFile, wsWord * word){
    
    Vector3D vec;
    
    // word
    *myFile << word->getString() << ",";
    // position
    word->getPosFinal(& vec);
    *myFile << vec[0] << "," << vec[1] << "," << vec[2] << ",";
    // color
    word->getCol(& vec);
    *myFile << vec[0] << "," << vec[1] << "," << vec[2] << ",";
    
    // synthesizer parameters
    for (int i = 7; i < NPARAMS-2; i++){
        switch (i) {
            case FMRATIO_0:
                *myFile << word->synthSettings->getFMRatios(0) << ",";
                break;
            case FMRATIO_1:
                *myFile << word->synthSettings->getFMRatios(1) << ",";
                break;
            case FMRATIO_2:
                *myFile << word->synthSettings->getFMRatios(2) << ",";
                break;
            case FMRATIO_3:
                *myFile << word->synthSettings->getFMRatios(3) << ",";
                break;
            case FMGAIN_0:
                *myFile << word->synthSettings->getFMGains(0) << ",";
                break;
            case FMGAIN_1:
                *myFile << word->synthSettings->getFMGains(1) << ",";
                break;
            case FMGAIN_2:
                *myFile << word->synthSettings->getFMGains(2) << ",";
                break;
            case FMGAIN_3:
                *myFile << word->synthSettings->getFMGains(3) << ",";
                break;
            case ADSRTIME_0:
                *myFile << word->synthSettings->getADSRTimes(0) << ",";
                break;
            case ADSRTIME_1:
                *myFile << word->synthSettings->getADSRTimes(1) << ",";
                break;
            case ADSRTIME_2:
                *myFile << word->synthSettings->getADSRTimes(2) << ",";
                break;
            case ASGAIN_0:
                *myFile << word->synthSettings->getASGain(0) << ",";
                break;
            case ASGAIN_1:
                *myFile << word->synthSettings->getASGain(1) << ",";
                break;
            case ASFREQ_0:
                *myFile << word->synthSettings->getASFreq(0) << ",";
                break;
            case ASFREQ_1:
                *myFile << word->synthSettings->getASFreq(1) << ",";
                break;
            case ASRAD_0:
                *myFile << word->synthSettings->getASRad(0) << ",";
                break;
            case ASRAD_1:
                *myFile << word->synthSettings->getASRad(1) << ",";
                break;
            case ECHO:
                *myFile << word->synthSettings->getEcho() << ",";
                break;
            case NOTE_0:
                *myFile << word->getFreqs(0) << ",";
                break;
            case NOTE_1:
                *myFile << word->getFreqs(1) << ",";
                break;
            case NOTE_2:
                *myFile << word->getFreqs(2) << ",";
                break;
            case NOTE_3:
                *myFile << word->getFreqs(3) << ",";
                break;
            case NOTE_4:
                *myFile << word->getFreqs(4) << ",";
                break;
            case TEMPO:
                *myFile << word->getTempo() << ",";
                
        }
    }
    
    // confidence & voting
    *myFile << word->getConfidence() << "," << word->getVotes() << endl;

    
}

// ---------------------
// update skew positions
// ---------------------
void Words::interp(GLfloat delta){
    for (auto& a: newWordsMap){
        a.second->interp(delta);
    }
    for (auto& a: editedWordsMap){
        a.second->interp(delta);
    }
    for (auto& a: existingWordsMap){
        a.second->interp(delta);
    }
}
void Words::setPosGoals(){
    
    for (auto& a: newWordsMap){
        a.second->setPosGoal();
    }
    for (auto& a: editedWordsMap){
        a.second->setPosGoal();
    }
    for (auto& a: existingWordsMap){
        a.second->setPosGoal();
    }
}
void Words::updatePositions(){
    
    for (auto& a: newWordsMap){
        a.second->updatePos();
    }
    for (auto& a: editedWordsMap){
        a.second->updatePos();
    }
    for (auto& a: existingWordsMap){
        a.second->updatePos();
    }
}

// ---
// PCA
// ---
void Words::resetPCA(){
    
    unsigned long nWords = existingWordsMap.size() + newWordsMap.size() + editedWordsMap.size();
    
    double content[nWords*(NPARAMS-15)];
    
    int i = 0;
    
    // get existing word parameters
    for (auto& a: existingWordsMap){
        for (int k = 0; k < NPARAMS-15; k++){
            content[i] = a.second->params[k];
            i++;
        }
    }
    // get new words parameters
    for (auto& a: newWordsMap){
        for (int k = 0; k < NPARAMS-15; k++){
            content[i] = a.second->params[k];
            i++;
        }
    }
    // get edited words parameters
    for (auto& a: editedWordsMap){
        for (int k = 0; k < NPARAMS-15; k++){
            content[i] = a.second->params[k];
            i++;
        }
    }
    
    // create algebra library array
    alglib::real_2d_array vals;
    vals.setcontent(nWords, NPARAMS-15, content);
    
    // create basis vectors
    alglib::ae_int_t info;
    alglib::ae_int_t &pInfo = info;
    alglib::pcabuildbasis(vals, nWords, NPARAMS-15, pInfo, pcaS2, pcaMapper);
    // the columns of pcaMapper contain basis vectors
    
    if ((info==1) && !(isinf(pcaS2[0]) & isinf(pcaS2[1]) & isinf(pcaS2[2])) ){
        pcaExists = 1;
        cerr << "resetPCA(), S2: " << pcaS2[0] << "," << pcaS2[1] << "," << pcaS2[2] << endl;
    }
    else pcaExists = 0;
    
}

// -------------------------------
// recompute all positions via PCA
// -------------------------------

void Words::recomputePos(){
    
    // recompute PCA
    resetPCA();
    
    if (pcaExists==1){
    
        // prepare reposition matrices
        alglib::real_1d_array params, result;
        alglib::real_1d_array &pParams = params;
        alglib::real_1d_array &pResult = result;
        alglib::real_2d_array &pMapper = pcaMapper;
    
        double zeros[NPARAMS-15];
        for (int i = 0 ; i < NPARAMS-15; i++) zeros[i] = 0;
    
        // reposition existing words
        for (auto& a: existingWordsMap){
            params.setcontent(NPARAMS-15, a.second->params);
            result.setcontent(NPARAMS-15, zeros);
            alglib::rmatrixmv(NPARAMS-15,NPARAMS-15,pMapper,0,0,1,pParams,0,pResult,0);
            a.second->setPosFinal(Vector3D(PCASCALE_0*result[1]/pcaS2[1],PCASCALE_1*result[2]/pcaS2[2],PCASCALE_2*result[3]/pcaS2[3]));
            a.second->setPosGoal(Vector3D(0,0,0));
        
        }
    
        // reposition edited words
        for (auto& a: editedWordsMap){
            params.setcontent(NPARAMS-15, a.second->params);
            result.setcontent(NPARAMS-15, zeros);
            alglib::rmatrixmv(NPARAMS-15,NPARAMS-15,pMapper,0,0,1,pParams,0,pResult,0);
            a.second->setPosFinal(Vector3D(PCASCALE_0*result[1]/pcaS2[1],PCASCALE_1*result[2]/pcaS2[2],PCASCALE_2*result[3]/pcaS2[3]));
            a.second->setPosGoal(Vector3D(0,0,0));
            
        }
    
        // reposition new words
        for (auto& a: newWordsMap){
            params.setcontent(NPARAMS-15, a.second->params);
            result.setcontent(NPARAMS-15, zeros);
            alglib::rmatrixmv(NPARAMS-15,NPARAMS-15,pMapper,0,0,1,pParams,0,pResult,0);
            a.second->setPosFinal(Vector3D(PCASCALE_0*result[1]/pcaS2[1],PCASCALE_1*result[2]/pcaS2[2],PCASCALE_2*result[3]/pcaS2[3]));
            a.second->setPosGoal(Vector3D(0,0,0));
            
        }
        
    }
    
}

// -----------------------------------
// Recompute a single position via PCA
// -----------------------------------

void Words::recomputePos(wsWord * w){
    
    if (pcaExists==1){
    
        double zeros[NPARAMS-15];
        for (int i = 0 ; i < NPARAMS-15; i++) zeros[i] = 0;
    
        // prepare reposition matrices
        alglib::real_1d_array params, result;
        alglib::real_1d_array &pParams = params;
        alglib::real_1d_array &pResult = result;
        alglib::real_2d_array &pMapper = pcaMapper;

        params.setcontent(NPARAMS-15, w->params);
        result.setcontent(NPARAMS-15, zeros);

        alglib::rmatrixmv(NPARAMS-15,NPARAMS-15,pMapper,0,0,1,pParams,0,pResult,0);

        w->setPosFinal(Vector3D(PCASCALE_0*result[1]/pcaS2[1],PCASCALE_1*result[2]/pcaS2[2],PCASCALE_2*result[3]/pcaS2[3]));
        w->setPosGoal();
        
        Vector3D pos;
        w->getPosFinal(&pos);
        cerr << "recomputePos: " << pos[0] << "," << pos[1] << "," << pos[2] << endl;
    }
    else{
        w->setPosFinal( Vector3D(10*(float)rand()/RAND_MAX-5,10*(float)rand()/RAND_MAX-5,10*(float)rand()/RAND_MAX-5) );
        w->setPosGoal();
    }
    
}

// ---------------------------
// save PCA parameters to file
// ---------------------------
void Words::savePCASettings(string fileName){
    
    if (pcaExists==1){
        ofstream myOut;
        myOut.open(Globals::tempFile);
        string myLine, myWord;
    
        for (int i = 0; i < NPARAMS-15; i++){
            for (int j = 0; j < NPARAMS-15; j++){
                myOut << pcaMapper[i][j] << ",";
            }
        }
        myOut << endl;
    
        for (int i = 0; i < NPARAMS-15; i++){
            myOut << pcaS2[i] << ",";
        }
        myOut << endl;
        myOut.close();
    
        // move temporary file to actual file
        rename( Globals::tempFile.c_str(), fileName.c_str() );
    }

}

// ------------------------------
// fetch PCA parameters from file
// ------------------------------
void Words::getPCASettings(string fileName){
    
    try{
        ifstream myIn;
        myIn.open(fileName.c_str());
        string myLine, myWord;
    
        pcaMapper.setlength(NPARAMS-15, NPARAMS-15);
    
        getline(myIn,myLine);
        int iChar = 0;
        int foundComma;
        for (int i = 0; i < NPARAMS-15; i++){
            for (int j = 0; j < NPARAMS-15; j++){
            
                foundComma = 0;
                myWord.clear();
                while (foundComma==0){
                    if (myLine[iChar]==',') foundComma=1;
                    else myWord.append(myLine.substr(iChar,1));
                
                    iChar++;
                }
                pcaMapper(i,j) = atof(myWord.c_str());
            
            }
        }
    
        pcaS2.setlength(NPARAMS-15);
        getline(myIn,myLine);
        iChar = 0;
        for (int i = 0; i < NPARAMS-15; i++){
            foundComma = 0;
            myWord.clear();
            while(foundComma==0){
                if (myLine[iChar]==',') foundComma=1;
                else myWord.append(myLine.substr(iChar,1));
            
                iChar++;
            }
            pcaS2(i) = atof(myWord.c_str());
        }
        
        pcaExists = 1;
        
    }
    catch (int e)
    {
        pcaExists = 0;
    }
    
}

// -------------------------------
// get distance squared from 0,0,0
// -------------------------------
float Words::getMaxDist(){
    
    float maxDist = 0;
    float dist;
    Vector3D *pos = new Vector3D();
    
    // reposition existing words
    for (auto& a: existingWordsMap){
        a.second->getPos(pos);
        dist = pow(pos->x,2)+pow(pos->y,2)+pow(pos->z,2);
        maxDist = (maxDist > dist) ? maxDist : dist;
    }
    
    // reposition edited words
    for (auto& a: editedWordsMap){
        a.second->getPos(pos);
        dist = pow(pos->x,2)+pow(pos->y,2)+pow(pos->z,2);
        maxDist = (maxDist > dist) ? maxDist : dist;
    }
    
    // reposition new words
    for (auto& a: newWordsMap){
        a.second->getPos(pos);
        dist = pow(pos->x,2)+pow(pos->y,2)+pow(pos->z,2);
        maxDist = (maxDist > dist) ? maxDist : dist;
    }

    return maxDist;
    
}

// ---------------------------------------
// get mean distance from finaly positions
// ---------------------------------------

float Words::getMeanDistFromFinal(){
    
    float meanDist = 0;
    float dist;
    Vector3D *pos = new Vector3D();
    Vector3D *posFinal = new Vector3D();
    
    // reposition existing words
    for (auto& a: existingWordsMap){
        a.second->getPos(pos);
        a.second->getPosFinal(posFinal);
        dist = pow(pos->x-posFinal->x,2)+pow(pos->y-posFinal->y,2)+pow(pos->z-posFinal->z,2);
        meanDist += dist;
    }
    
    // reposition edited words
    for (auto& a: editedWordsMap){
        a.second->getPos(pos);
        a.second->getPosFinal(posFinal);
        dist = pow(pos->x-posFinal->x,2)+pow(pos->y-posFinal->y,2)+pow(pos->z-posFinal->z,2);
        meanDist += dist;
    }
    
    // reposition new words
    for (auto& a: newWordsMap){
        a.second->getPos(pos);
        a.second->getPosFinal(posFinal);
        dist = pow(pos->x-posFinal->x,2)+pow(pos->y-posFinal->y,2)+pow(pos->z-posFinal->z,2);
        meanDist += dist;
    }
    
    meanDist /= nWords;
    return meanDist;
    
}


// ---------------
// reset randNotes
// ---------------
void Words::resetRandFreqs(){
    
    for (auto& a: existingWordsMap){
        a.second->setRandFreq(Globals::pastFreqs, NPASTFREQS);
    }
    for (auto& a: editedWordsMap){
        a.second->setRandFreq(Globals::pastFreqs, NPASTFREQS);
    }
    for (auto& a: newWordsMap){
        a.second->setRandFreq(Globals::pastFreqs, NPASTFREQS);
    }
    
}


// ---------------------------
// initialize: wsSynthSettings
// ---------------------------

void wsSynthSettings::setSynth(wsSynth * syn){
        
    // fm
    for (int i = 0; i < NCARRIERS; i++){
        float val = pow(fmRatios_[i],.5);
        syn->fm.setRatio(i, StkFloat(val) );
        syn->fm.setGain(i, fmGains_[i]);
    }
    syn->fm.setRatio(NCARRIERS, fmRatios_[NCARRIERS]);
    syn->fm.setGain(NCARRIERS, pow(fmGains_[NCARRIERS],4));
    
    // adsr attack targets
    syn->adsrGain.setAttackTarget(asGain_[0]);
    syn->adsrFreq.setAttackTarget(pow(10, asFreq_[0]));
    syn->adsrRad.setAttackTarget(asRad_[0]);
    // adsr sustain targets & attack, decay, release times
    syn->adsrGain.setAllTimes(adsrTimes_[0],adsrTimes_[1],asGain_[1],adsrTimes_[2]);
    syn->adsrFreq.setAllTimes(adsrTimes_[0],adsrTimes_[1],asFreq_[1],adsrTimes_[2]);
    syn->adsrRad.setAllTimes(adsrTimes_[0],adsrTimes_[1],asRad_[1],adsrTimes_[2]);
    
    // echo
    syn->echo.setDelay((unsigned long) echo_ );
    
};

// --------------------------
// randomize: wsSynthSettings
// --------------------------

void wsSynthSettings::randomize(){
    
    // fm carriers
    fmRatios_[0] = 1;
    fmGains_[0] = .6;
    for (int i = 1; i < NCARRIERS; i++){
        fmRatios_[i] = fmRatioMin + (fmRatioMax - fmRatioMin)*float(rand())/RAND_MAX;
        fmGains_[i] = fmGainMin + (fmGainMax-fmGainMin)*float(rand())/RAND_MAX;
    }
    // fm modulator
    fmRatios_[NCARRIERS-1] = fmModRatioMin + (fmModRatioMax-fmModRatioMin)*float(rand())/RAND_MAX;
    fmGains_[NCARRIERS-1] = fmModGainMin + (fmModGainMax-fmModGainMin)*float(rand())/RAND_MAX;
    
    // adsr times
    adsrTimes_[0] = attackMin + (attackMax-attackMin)*float(rand())/RAND_MAX;
    adsrTimes_[1] = decayMin + (decayMax-decayMin)*float(rand())/RAND_MAX;
    adsrTimes_[2] = releaseMin + (releaseMax-releaseMin)*float(rand())/RAND_MAX;
    
    // adsr sustains
    asGain_[0] = attackGainMin + (attackGainMax-attackGainMin)*float(rand())/RAND_MAX;
    asGain_[1] = susGainMin + (asGain_[0]-susGainMin)*float(rand())/RAND_MAX;
    asRad_[0] = attackRadMin + (attackRadMax-attackRadMin)*float(rand())/RAND_MAX;
    asRad_[1] = susRadMinRatio*asRad_[0] + (susRadMax-susRadMinRatio*asRad_[0])*float(rand())/RAND_MAX;
    asFreq_[0] = attackFreqMin + (attackFreqMax-attackFreqMin)*float(rand())/RAND_MAX;
    asFreq_[1] = susFreqMin + (asFreq_[0]-susFreqMin)*float(rand())/RAND_MAX;
    
    // echo
    echo_ = (unsigned long) .8*SRATE*(echoMin + (echoMax-echoMin)*float(rand())/RAND_MAX);

    
}

// --------------------------
// randomize: wsSynthSettings
// --------------------------

void wsSynthSettings::randomize(float confidence){
    
    StkFloat mult = 1-pow(confidence, .5);

    StkFloat min, max;
    
    // fm carriers
    fmRatios_[0] = 1;
    fmGains_[0] = .8;
    for (int i = 1; i < NCARRIERS; i++){
        min = fmRatios_[i]-(fmRatios_[i]-fmRatioMin)*mult;
        max = fmRatios_[i]-(fmRatios_[i]-fmRatioMax)*mult;
        fmRatios_[i] = min + (max-min)*float(rand())/RAND_MAX;
        min = fmGains_[i]-(fmGains_[i]-fmGainMin)*mult;
        max = fmGains_[i]-(fmGains_[i]-fmGainMax)*mult;
        fmGains_[i] = min + (max-min)*float(rand())/RAND_MAX;
    }
    
    // fm modulator
    min = fmRatios_[NCARRIERS]-(fmRatios_[NCARRIERS]-fmModRatioMin)*mult;
    max = fmRatios_[NCARRIERS]-(fmRatios_[NCARRIERS]-fmModRatioMax)*mult;
    fmRatios_[NCARRIERS] = min + (max-min)*float(rand())/RAND_MAX;
    min = fmGains_[NCARRIERS]-(fmGains_[NCARRIERS]-fmModGainMin)*mult;
    max = fmGains_[NCARRIERS]-(fmGains_[NCARRIERS]-fmModGainMax)*mult;
    fmGains_[NCARRIERS] = min + (max-min)*float(rand())/RAND_MAX;
    
    // adsr times
    min = adsrTimes_[0]-(adsrTimes_[0]-attackMin)*mult;
    max = adsrTimes_[0]-(adsrTimes_[0]-attackMax)*mult;
    adsrTimes_[0] = min + (max-min)*float(rand())/RAND_MAX;
    
    min = adsrTimes_[1]-(adsrTimes_[1]-decayMin)*mult;
    max = adsrTimes_[1]-(adsrTimes_[1]-decayMax)*mult;
    adsrTimes_[1] = min + (max-min)*float(rand())/RAND_MAX;
    
    min = adsrTimes_[2]-(adsrTimes_[2]-releaseMin)*mult;
    max = adsrTimes_[2]-(adsrTimes_[2]-releaseMax)*mult;
    adsrTimes_[2] = min + (max-min)*float(rand())/RAND_MAX;
    
    // adsr sustains
    min = asGain_[0]-(asGain_[0]-attackGainMin)*mult;
    max = asGain_[0]-(asGain_[0]-attackGainMax)*mult;
    asGain_[0] = min + (max-min)*float(rand())/RAND_MAX;
    min = asGain_[1]-(asGain_[1]-susGainMin)*mult;
    max = asGain_[1]-(asGain_[1]-asGain_[0])*mult;
    asGain_[1] = min + (max-min)*float(rand())/RAND_MAX;
    
    min = asRad_[0]-(asRad_[0]-attackRadMin)*mult;
    max = asRad_[0]-(asRad_[0]-attackRadMax)*mult;
    asRad_[0] = min + (max-min)*float(rand())/RAND_MAX;
    min = asRad_[1]-(asRad_[1]-asRad_[0]*susRadMinRatio)*mult;
    max = asRad_[1]-(asRad_[1]-susRadMax)*mult;
    asRad_[1] = min + (max-min)*float(rand())/RAND_MAX;

    min = asFreq_[0]-(asFreq_[0]-attackFreqMin)*mult;
    max = asFreq_[0]-(asFreq_[0]-attackFreqMax)*mult;
    asFreq_[0] = min + (max-min)*float(rand())/RAND_MAX;
    min = asFreq_[1]-(asFreq_[1]-susFreqMin)*mult;
    max = asFreq_[1]-(asFreq_[1]-asFreq_[0])*mult;
    asFreq_[1] = min + (max-min)*float(rand())/RAND_MAX;
    
    // echo
    min = echo_-(echo_-echoMin)*mult;
    max = echo_-(echo_-echoMax)*mult;
    echo_ = (unsigned long) .8*SRATE*(min + (max-min)*float(rand())/RAND_MAX);

    
}

// ---------------------------------------------
// setParams: wsWord [from synthesizer settings]
// ---------------------------------------------
void wsWord::setParams(wsSynthSettings * s){
    
    for (int i = 7; i < NPARAMS-2; i++){
        switch (i) {
            case FMRATIO_0:
                params[FMRATIO_0-7] = s->getFMRatios(0);
                break;
            case FMRATIO_1:
                params[FMRATIO_1-7] = s->getFMRatios(1);
                break;
            case FMRATIO_2:
                params[FMRATIO_2-7] = s->getFMRatios(2);
                break;
            case FMRATIO_3:
                params[FMRATIO_3-7] = PARAMS_SCALE_FMRATIO_3*s->getFMRatios(3);
                break;
            case FMGAIN_0:
                params[FMGAIN_0-7] = s->getFMGains(0);
                break;
            case FMGAIN_1:
                params[FMGAIN_1-7] = s->getFMGains(1);
                break;
            case FMGAIN_2:
                params[FMGAIN_2-7] = s->getFMGains(2);
                break;
            case FMGAIN_3:
                params[FMGAIN_3-7] = PARAMS_SCALE_FMGAIN_3*s->getFMGains(3);
                break;
            case ADSRTIME_0:
                params[ADSRTIME_0-7] = s->getADSRTimes(0);
                break;
            case ADSRTIME_1:
                params[ADSRTIME_1-7] = s->getADSRTimes(1);
                break;
            case ADSRTIME_2:
                params[ADSRTIME_2-7] = s->getADSRTimes(2);
                break;
            case ASGAIN_0:
                params[ASGAIN_0-7] = s->getASGain(0);
                break;
            case ASGAIN_1:
                params[ASGAIN_1-7] = s->getASGain(1);
                break;
            case ASFREQ_0:
                params[ASFREQ_0-7] = s->getASFreq(0);
                break;
            case ASFREQ_1:
                params[ASFREQ_1-7] = s->getASFreq(1);
                break;
            case ASRAD_0:
                params[ASRAD_0-7] = s->getASRad(0);
                break;
            case ASRAD_1:
                params[ASRAD_1-7] = PARAMS_SCALE_ASRAD_1*s->getASRad(1);
                break;
            case ECHO:
                params[ECHO-7] = s->getEcho();
                break;

        }
    }
    
}
