#define fmRatioMin 1/16
#define fmRatioMax 50
#define fmGainMin 0
#define fmGainMax .3
#define fmModRatioMin .25
#define fmModRatioMax 8
#define fmModGainMin .01
#define fmModGainMax 1.4
#define attackMin .1
#define attackMax .5
#define decayMin .05
#define decayMax .45
#define releaseMin .05
#define releaseMax .25
#define attackGainMin .5
#define attackGainMax .8
#define susGainMin .2
#define attackRadMin .1
#define attackRadMax .95
#define susRadMinRatio .5
#define susRadMax .95
#define attackFreqMin 0
#define attackFreqMax 2
#define susFreqMin 0
#define echoMin 0
#define echoMax 1

#define NCARRIERS 3

#define NPARAMS 33

#define NPASTFREQS 5

#define RANDNOTE_MIN 24
#define RANDNOTE_MAX 84
#define RANDNOTE_DIF 2

#ifndef __W_ENTITY_H__
#define __W_ENTITY_H__

#include <vector>
#include "y-entity.h"
#include <ioStream>
#include <unordered_map>

#include "Stk.h"
#include "BandedWG.h"
#include "SineWave.h"

#include "synth.h"

#include "globals.h"
#include "sim.h"

#include "linalg.h"
#include "ap.h"
#include "dataanalysis.h"

using namespace std;
using namespace stk;

// forward reference
class wsSynth;
class Globals;
class Sim;

// ----------------------
// class: wsSynthSettings
// ----------------------
class wsSynthSettings{
    
private:
    
    int alreadySet;
    
    StkFloat * fmRatios_;
    StkFloat * fmGains_;
    
    StkFloat * adsrTimes_;
    
    StkFloat * asGain_;
    StkFloat * asFreq_;
    StkFloat * asRad_;
    
    StkFloat echo_;
    
public:
    
    wsSynthSettings(){
        fmRatios_ = (StkFloat *) malloc((NCARRIERS+1)*sizeof(StkFloat));
        fmGains_ = (StkFloat *) malloc((NCARRIERS+1)*sizeof(StkFloat));
        adsrTimes_ = (StkFloat *) malloc(3*sizeof(StkFloat));
        asGain_ = (StkFloat *) malloc(2*sizeof(StkFloat));
        asFreq_ = (StkFloat *) malloc(2*sizeof(StkFloat));
        asRad_ = (StkFloat *) malloc(2*sizeof(StkFloat));
        
        for (int i = 0; i < NCARRIERS+1; i++){
            fmRatios_[i] = 0;
            fmGains_[i] = 1;
        }
        
        for (int i = 0; i < 3; i++) adsrTimes_[i] = 0;
        for (int i = 0; i < 2; i++){
            asGain_[i] = 0;
            asFreq_[i] = 0;
            asRad_[i] = 0;
        }
        
        echo_ = 0;
        
    };
    ~wsSynthSettings(){};
    
    void randomize();
    
    void randomize(float confidence);
    
    void setSynth(wsSynth * syn);
    
    // sets
    void setFMRatios(int i, StkFloat val){ fmRatios_[i] = val; };
    void setFMGains(int i, StkFloat val){ fmGains_[i] = val; };
    void setADSRTimes(int i, StkFloat val){ adsrTimes_[i] = val; };
    void setASGain(int i, StkFloat val){ asGain_[i] = val; };
    void setASFreq(int i, StkFloat val){ asFreq_[i] = val; };
    void setASRad(int i, StkFloat val){ asRad_[i] = val; };
    void setEcho(StkFloat val){ echo_ = val; };
    
    // gets
    StkFloat getFMRatios(int i){ return fmRatios_[i]; };
    StkFloat getFMGains(int i){ return fmGains_[i]; };
    StkFloat getADSRTimes(int i){ return adsrTimes_[i]; };
    StkFloat getASGain(int i){ return asGain_[i]; };
    StkFloat getASFreq(int i){ return asFreq_[i]; };
    StkFloat getASRad(int i){ return asRad_[i]; };
    unsigned long getEcho(){ return echo_; };
    
    // updates
    void updateSynthSettings(wsSynthSettings * syn){
        
        // fm carriers
        for (int i = 0; i < NCARRIERS; i++){
            fmRatios_[i] = syn->getFMRatios(i);
            fmGains_[i] = syn->getFMGains(i);
        }
        // fm modulator
        fmRatios_[NCARRIERS-1] = syn->getFMRatios(NCARRIERS-1);
        fmGains_[NCARRIERS-1] = syn->getFMGains(NCARRIERS-1);
        
        // adsr times
        for (int i = 0; i < 3; i++) adsrTimes_[i] = syn->getADSRTimes(i);
        
        // adsr sustains
        for (int i = 0; i < 2; i++){
            asGain_[i] = syn->getASGain(i);
            asFreq_[i] = syn->getASFreq(i);
            asRad_[i] = syn->getASRad(i);
        }
        
        // echo
        echo_ = syn->getEcho();

        
    };
    
};

// -------------
// class: wsWord
// -------------
class wsWord{
    
private:
    
    string strText = "";
    int strLen = 0;
    
    Vector3D posX = Vector3D(0,0,SLEWRATE_WORDS);
    Vector3D posY = Vector3D(0,0,SLEWRATE_WORDS);
    Vector3D posZ = Vector3D(0,0,SLEWRATE_WORDS);

    Vector3D position = Vector3D(0,0,0);

    Vector3D color = Vector3D(0,0,0);
    
    float confidenceNew, confidenceUpdated;
    int confidenceVotes, confidenceVoted;
    
    float randFreq;
    
    float pastFreqs[NPASTFREQS];
    float eighthNoteTempo;
    
public:
    
    wsSynthSettings * synthSettings;
    wsSynthSettings * oldSynthSettings;
    YText * yText;
    double params[NPARAMS-9];
    
    // constructor/destructor
    wsWord(){
        yText = new YText();
        yText->active = true;
        yText->sca.set(1,1,1);
        synthSettings = new wsSynthSettings();
        oldSynthSettings = new wsSynthSettings();
        eighthNoteTempo = 120;
        for (int i = 0; i < NPASTFREQS; i++) pastFreqs[i] = 880;
    };
    ~wsWord(){};
    
    // set
    void setString(string txt, int len){
        strText = txt.c_str();
        yText->set(strText);
        
    };
    void setPosFinal(Vector3D pos){position = pos;};
    
    void setPosNow(Vector3D pos){ posX.value = pos.x; posY.value = pos.y; posZ.value = pos.z; updatePos(); };
    void setPosNow(){ posX.value=position.x; posY.value=position.y; posZ.value=position.z; updatePos(); };
    void setPosGoal(Vector3D pos){ posX.goal=pos.x; posY.goal=pos.y; posZ.goal=pos.z; };
    void setPosGoal(){ posX.goal=position.x; posY.goal=position.y; posZ.goal=position.z; };
    
    void interp(GLfloat delta){ posX.interp(delta); posY.interp(delta); posZ.interp(delta); };
    void updatePos(){yText->setCenterLocation(Vector3D(posX.value,posY.value,posZ.value));};
    void updatePos(Vector3D trans){yText->setCenterLocation(Vector3D(posX.value+trans.x,posY.value+trans.y,posZ.value+trans.z));};
    void updatePos(Vector3D trans, float val){yText->setCenterLocation(Vector3D(posX.value+val*trans.x,posY.value+val*trans.y,posZ.value+val*trans.z));};

    
    void setColor(Vector3D col){color = col; yText->setColor(Vector3D(color[0]+confidenceUpdated*(1-color[0]),color[1]+confidenceUpdated*(1-color[1]),color[2]+confidenceUpdated*(1-color[2])));};
    void setSynth(wsSynthSettings * syn){ synthSettings->updateSynthSettings(syn); setParams(synthSettings); };
    void setParams(double * p){for (int i = 0; i < NPARAMS-9; i++) params[i] = p[i]; }
    void setParams(wsSynthSettings * s);
    
    void setRandFreq(float val){randFreq = val;};
    void setRandFreq(GLfloat * vals, int nVals){
        int ind = floor( nVals*float(rand())/RAND_MAX );
        int power = floor( 1*float(rand())/RAND_MAX );
        
        randFreq = vals[ind] * pow(2,power);
        
        while(randFreq > 15000) randFreq = randFreq/2;
        while(randFreq < 400) randFreq = randFreq * 2;
    };
    
    void setFreqs(int i, float val){ pastFreqs[i] = val; }
    void setTempo(float val){ eighthNoteTempo = val; }
    
    // get
    string getString(){
        return strText;
    }
    void getPos(Vector3D * pos){ pos->set(posX.value, posY.value, posZ.value); };
    void getPosFinal(Vector3D * pos){pos->set(position.x, position.y, position.z); };
    void getCol(Vector3D * col){ col->set(color[0], color[1], color[2]); }
    
    float getRandFreq(){ return randFreq; };
    
    float getFreqs(int i){ return pastFreqs[i]; }
    
    float getTempo(){return eighthNoteTempo; }
    
    void copy(wsWord wsCopy){
        wsCopy.setString(strText, strLen);
        wsCopy.setPosFinal(position);
        wsCopy.setPosNow();
        wsCopy.setColor(color);
    }
    
    void duplicate(string str, Vector3D pos, Vector3D col){
        this->setString(str, (int) str.length());
        this->setPosFinal(pos);
        this->setPosNow();
        this->setColor(col);
    }
    
    // synthesize
    void randomizeSynthSettings(){ synthSettings->randomize(); setParams(synthSettings); };
    void randomizeSynthSettings(float conf){ synthSettings->randomize(conf); setParams(synthSettings); };
    void setSynth(wsSynth * syn){ synthSettings->setSynth(syn); };

    // confidence
    void initializeConfidence(){ confidenceVoted = 0; oldSynthSettings->updateSynthSettings(synthSettings); synthSettings->randomize(confidenceUpdated); setParams(synthSettings); }
    void setConfidence(float val){ confidenceNew = val; confidenceVoted = 1; };
    void resetConfidence(float val, int n){ confidenceUpdated = val; confidenceVotes = n; };
    float getConfidence(){ return confidenceUpdated; };
    int getVotes(){ return confidenceVotes; };
    void updateConfidence(){
        float lastConfidence = confidenceUpdated;
        cerr << "number of previous votes: " << confidenceVotes << endl;
        cerr << "previous confidence: " << confidenceUpdated << endl;
        // update confidence
        if (confidenceVoted == 1){
            cerr << "recent vote: " << confidenceNew << endl;
            confidenceUpdated = (float(confidenceVotes)*confidenceUpdated + confidenceNew)/float(confidenceVotes + 1);
            cerr << "updated confidence: " << confidenceUpdated << endl;
            confidenceVotes++;
        }
        // update root synthesizer settings
        if (lastConfidence <= confidenceUpdated) oldSynthSettings->updateSynthSettings(synthSettings);
    }
    void updateConfidence(wsSynthSettings * settings, float conf){
        float lastConfidence = confidenceUpdated;
        // specify updated synth settings
        synthSettings->updateSynthSettings(settings);
        // update confidence
        confidenceUpdated = (float(confidenceVotes)*confidenceUpdated + conf)/float(confidenceVotes + 1);
        confidenceVotes++;
        // update synthesizer settings
        if(lastConfidence <= confidenceUpdated) oldSynthSettings->updateSynthSettings(synthSettings);
    }

};

// ------------
// class: Words
// ------------
class Words {

private:
    
    alglib::real_2d_array pcaMapper;
    alglib::real_1d_array pcaS2;
    int pcaExists = 0;
    int nWords = 0;
    
public:

    // hash map of words
    unordered_map<string, wsWord *> existingWordsMap, newWordsMap, editedWordsMap;

    // constructor/destructor
    Words(){};
    ~Words(){};
    
    // initialize
    void initialize(string fileName);
    
    // save session
    void save(string fileName);
    void saveLine(ofstream * myFile, wsWord * word);
    
    // add new word
    void addNew(wsWord * word){
        wsWord * newWord = new wsWord();
        Vector3D pos, col;
        word->getPosFinal(&pos);
        word->getCol(&col);
        newWord->setSynth( word->synthSettings );
        newWord->duplicate(word->getString(), pos, col);
        newWord->resetConfidence(word->getConfidence(), word->getVotes());
        newWordsMap.insert({{newWord->getString(), newWord}});
    };
    
    // add list
    void addList(string, wsWord * baseWord);
    
    // is listed
    int isListedExistingWordsMap(string str){
        unordered_map<string,wsWord *>::const_iterator got = existingWordsMap.find(str);
        if (got!= existingWordsMap.end()) return 1;
        else return 0;
    }
    int isListedNewWordsMap(string str){
        unordered_map<string,wsWord *>::const_iterator got = newWordsMap.find(str);
        if (got!= newWordsMap.end()) return 1;
        else return 0;
    }
    int isListedEditedWordsMap(string str){
        unordered_map<string,wsWord *>::const_iterator got = editedWordsMap.find(str);
        if (got!= editedWordsMap.end()) return 1;
        else return 0;
    }
    int isListed(string str){
        
        if (isListedExistingWordsMap(str)==1) return 1;
        if (isListedNewWordsMap(str)==1) return 2;
        if (isListedEditedWordsMap(str)==1) return 3;
        else return 0;
    }
    
    // update positions
    void interp(GLfloat delta);
    void setPosGoals();
    void updatePositions();
    
    // PCA
    void resetPCA();
    void recomputePos();
    void recomputePos(wsWord * w);
    void savePCASettings(string fileName);
    void getPCASettings(string fileName);
    
    // get distance-squared from 0
    float getMaxDist();
    float getMeanDistFromFinal();
    
    // reset random notes
    void resetRandFreqs();
    
    // get number of words
    int getNWords(){ return nWords; };
    
};

#endif